@@ -13,23 +13,20 @@ import yaml
 from concurrent.futures import ThreadPoolExecutor
 from queue import Queue
 from urllib3.exceptions import InsecureRequestWarning
+import logging
 
-from difflib import context_diff
 
 
 requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)
 
-#during multithread
+logging.basicConfig(format=' %(message)s')
 
 class dyagram:
 
-    # cdp_info = {} will need global list in future when multithreaded/multiprocessed
-    # Right now class only supports SSH (netmiko), will need to enable REST later
+    def __init__(self, inventory_file=None, verbose=False):
 
-    def __init__(self, inventory_file=None, new_state=None):
-
-        self.state_exists = self.does_state_exist()
         self.inventory_file = inventory_file
+        self.verbose = verbose
         self.inventory_object = self.get_inv_yaml_obj()
         self._devices_to_query = Queue()
         self._devices_queried = []
@@ -37,6 +34,7 @@ class dyagram:
         self.username = None
         self.password = None
         self.site = self.get_current_site()
+        self.state_exists = self.does_state_exist()
         self._neighbor_template = {
             "hostname": "",
             "local_port": "",
@@ -46,13 +44,27 @@ class dyagram:
 
         self._load_creds()
         self._load_inventory()
+        self.log = None
+        self.setup_logging()
+
+
+    def setup_logging(self):
+        log = logging.getLogger("DYAGRAM")
+        log.setLevel(logging.INFO)
+        self.log = log
+        if self.verbose:
+            self.log.setLevel(logging.INFO)
+        else:
+            self.log.setLevel(logging.NOTSET)
 
 
     def get_current_site(self):
         try:
+
             file = open(r".info/info.json", 'r')
             info = json.load(file)
             #print(info['current_site'])
+            file.close()
             return info['current_site']
         except:
             print("Unable to load last site.")
@@ -65,50 +77,70 @@ class dyagram:
 
     def _load_creds(self):
         #first try OS ENV
+        self.log.info("Loading Credentials")
         self.username = os.environ['DYAGRAM_USERNAME']
         self.password = os.environ['DYAGRAM_PASSWORD']
+        self.log.info("Loaded Credentials")
 
-    def __discover_restconf_ssh(self, device):
-        '''
+
+    def __discover_lldp_neighbors(self, device):
+
+        """
         this enables us to keep multithreading and not block for waiting for an exception for restconf before trying
         ssh.
 
         See inside discover() for threading
         :param device:
         :return:
-        '''
+        """
 
         try:
-            resp = self._discover_neighbors_by_restconf(device)
-            if not resp:
+            self.log.info(f"DEVICE: {device} - RESTCONF : Discovering LLDP Neighbors - START")
+            resp = self._discover_lldp_neighbors_by_restconf(device)
+            if not resp or resp.status_code != 200:
                 raise
+            self.log.info(f"DEVICE: {device} - RESTCONF : Discovering LLDP Neighbors - SUCCESSFUL")
         except:
             # try ssh
-            self._discover_neighbors_by_ssh(device)
+            self.log.info(f"DEVICE: {device} - RESTCONF : Discovering LLDP Neighbors - FAILURE")
+            self.log.info(f"DEVICE: {device} - SSH : Discovering LLDP Neighbors - START")
+            self._discover_lldp_neighbors_by_ssh(device)
+            self.log.info(f"DEVICE: {device} - SSH : Discovering LLDP Neighbors - SUCCESSFUL")
+
 
     def discover(self):
-        '''
+
+        """
 
         Generates topology via cdp/lldp neighbors into a json object assigned to topology attribute
 
         :return:
-        '''
+        """
 
-
-        executor = ThreadPoolExecutor(max_workers=10)
+        executor = ThreadPoolExecutor(max_workers=30)
 
         queue_empty = False
         while not queue_empty:
             try:
-               device =  self._devices_to_query.get_nowait()
-               executor.submit(self.__discover_restconf_ssh, device)
-               self._devices_queried.append(device)
+                device = self._devices_to_query.get_nowait()
+                self.log.info(f"Pulled device {device} from Queue")
+                self.topology['devices'].append({'hostname': "", 'inventory_ip': device, 'layer2': {}, 'routes': []})
+
+                executor.submit(self.__discover_lldp_neighbors, device)
+                self.log.info(f"DEVICE: {device} - Submitted for discover_lldp_neighbors")
+                executor.submit(self.discover_routes, device)
+                self.log.info(f"DEVICE: {device} - Submitted for discover_routes")
+                #executor.submit(self.__discover_dynamic_routing_neighbors, device)
+
+                self._devices_queried.append(device)
 
             except queue.Empty:
+                self.log.info(f"QUEUE EMPTY")
                 queue_empty = True
             except Exception:
                 tb = self.get_traceback()
-                #print(tb)
+                self.log.info(f"EXCEPTION THROWN IN PULLING FROM QUEUE: {tb}")
+
 
         executor.shutdown(wait=True)
 
@@ -116,14 +148,13 @@ class dyagram:
 
         self.topology['devices'] = sorted(self.topology['devices'], key=lambda d: d['hostname'])
 
-
         if not self.state_exists:
             self.export_state()
         else:
             self.compare_states()
 
     def does_state_exist(self):
-        path = Path("lab/state.json")
+        path = Path(f"{self.site}/state.json")
         return path.is_file()
 
     def export_state(self):
@@ -133,22 +164,118 @@ class dyagram:
         file.close()
 
     def compare_states(self):
+
         file = open(rf"{self.site}/state.json", 'r')
         state = json.load(file)
         current_state = self.topology
 
         if state == current_state:
-            print("--DyaGram Discovery Complete--")
-            print("\n\nNo changes in state\n\n")
+            print("\nNo changes in state\n")
 
         else:
-            print("\n\n--DyaGram Discovery Complete--")
             print("\n\nCHANGES IN STATE!!")
             print(f"\n\nSTATE FILE:\n{state}")
             print(f"\n\nCURRENT STATE:\n{current_state}")
         file.close()
 
 
+    def discover_routes(self, device):
+
+        try:
+            self.log.info(f"DEVICE: {device} - RESTCONF : Discovering routes - START")
+            routes = self.discover_routes_restconf(device)
+            if not routes or routes.status_code != 200:
+                raise
+            self.log.info(f"DEVICE: {device} - RESTCONF : Discovering routes - SUCCESSFUL")
+        except:
+            self.log.info(f"DEVICE: {device} - RESTCONF : Discovering routes - FAILURE")
+            self.log.info(f"DEVICE: {device} - SSH : Discovering routes - START")
+            routes = self.discover_routes_ssh(device)
+            self.log.info(f"DEVICE: {device} - SSH : Discovering routes - SUCCESSFUL")
+
+        return True
+
+    def discover_routes_restconf(self):
+        return None
+
+    def discover_routes_ssh(self, device):
+
+        CONN_SET = False
+
+        autodetect_netmiko_args = {"device_type": "autodetect",
+                                   "host": device,
+                                   "username": self.username,
+                                   "password": self.password}
+
+        netmiko_args = {"device_type": "",
+                        "host": device,
+                        "username": self.username,
+                        "password": self.password,
+                        "secret": self.password}
+
+        try:
+            guesser = SSHDetect(**autodetect_netmiko_args)
+            best_match = guesser.autodetect()
+            netmiko_args['device_type'] = best_match
+
+        except NetmikoAuthenticationException:
+            # print(f"AUTHENTICATION ERROR FOR DEVICE: {device}")
+            return False
+        except:
+            try:
+                if not netmiko_args['device_type']:
+                    # Later put in unable to find OS
+                    netmiko_args['device_type'] = "cisco_ios"
+                dev = ConnectHandler(**netmiko_args)
+                dev.enable()
+            except Exception:
+                tb = self.get_traceback()
+                # print(tb)
+
+            CONN_SET = True
+            os = self._get_os_version(dev)
+            netmiko_args['device_type'] = os
+
+        if not CONN_SET:
+            dev = ConnectHandler(**netmiko_args)
+
+        dev.enable()
+
+        routes = dev.send_command("show ip route vrf all", use_textfsm=True)
+
+        for route in routes:
+            if 'uptime' in route.keys():
+                route.pop('uptime')
+
+        dev.disconnect()
+        for i in self.topology['devices']:
+            if i['inventory_ip'] == device:
+                i['routes'] = routes
+                break
+
+    def __discover_dynamic_routing_neighbors(self, device):
+
+        self.discover_ospf_neighbors(device)
+
+        self.discover_eigrp_neigbors(device)
+
+        self.discover_bgp_neighbors(device)
+
+    def discover_ospf_neighbors(self, device):
+        pass
+
+    def discover_eigrp_neigbors(self, device):
+        pass
+
+    def discover_bgp_neighbors(self, device):
+        pass
+
+    def discover_ospf_neighbors_via_ssh(self, session):
+
+        pass
+
+    def discover_ospf_neighbors_via_restconf(self):
+        pass
 
     def get_inv_yaml_obj(self):
 
@@ -175,10 +302,16 @@ class dyagram:
             self._devices_to_query.put(ip)
 
 
-    def _discover_neighbors_by_restconf(self, device):
+    def _discover_lldp_neighbors_by_restconf(self, device):
         try:
             lldp_neighbors = self._get_lldp_neighbors_restconf(device)
-            self.topology["devices"].append(lldp_neighbors)
+
+            for i in self.topology['devices']:
+                if i['inventory_ip'] == device:
+                    i['hostname'] = lldp_neighbors['hostname']
+                    lldp_neighbors.pop('hostname')
+                    i['layer2'] = lldp_neighbors
+                    break
 
             self._devices_queried.append(device)
         except:
@@ -186,7 +319,7 @@ class dyagram:
 
         return True
 
-    def _discover_neighbors_by_ssh(self, device):
+    def _discover_lldp_neighbors_by_ssh(self, device):
 
         CONN_SET = False
 
@@ -225,7 +358,8 @@ class dyagram:
 
         if not CONN_SET:
             dev = ConnectHandler(**netmiko_args)
-            dev.enable()
+
+        dev.enable()
 
 
         try:
@@ -240,9 +374,12 @@ class dyagram:
             tb = self.get_traceback()
             #print(tb)
 
-
-        self.topology["devices"].append(lldp_nei_json)
-
+        for i in self.topology["devices"]:
+            if i['inventory_ip'] == device:
+                i['hostname'] = lldp_nei_json['hostname']
+                lldp_nei_json.pop("hostname")
+                i['layer2'].append(lldp_nei_json)
+                break
 
         dev.disconnect()
         self._devices_queried.append(device)
@@ -261,7 +398,7 @@ class dyagram:
                 resp = restconf_session.get(f"{restconf_session.base_url}/openconfig-interfaces:interfaces")
             except Exception:
                 tb = self.get_traceback()
-                #print(tb)
+                #print(tb) LOG THIS
 
             if resp.status_code == 200:
                 return  [i['ethernet']['state']['hw-mac-address']
@@ -316,9 +453,12 @@ class dyagram:
         session.base_url = f"https://{device}/restconf/data"
 
         #first try OpenConfig YANG
-
+        self.log.info(f"DEVICE: {device} - RESTCONF-OPENCONFIG : Querying lldp for device - START")
         oc_yang_resp = session.get(f"{session.base_url}/openconfig-lldp:lldp/interfaces/")
-        oc_yang_resp.raise_for_status()
+        if oc_yang_resp.status_code != 200:
+            self.log.info(f"DEVICE: {device} - RESTCONF-OPENCONFIG : Querying lldp for device - FAILURE")
+            oc_yang_resp.raise_for_status()
+        self.log.info(f"DEVICE: {device} - RESTCONF-OPENCONFIG : Querying lldp for device - SUCCESSFUL")
 
 
 
@@ -445,26 +585,38 @@ class dyagram:
 
         if not netmiko_session and restconf_session:
             try:
+                self.log.info(f"DEVICE: {restconf_session.base_url} - RESTCONF_OPENCONFIG : Querying for hostname - START")
                 resp = restconf_session.get(f"{restconf_session.base_url}/openconfig-system:system/config/name")
                 if resp.status_code == 200:
+                    self.log.info(f"DEVICE: {restconf_session.base_url} - RESTCONF_OPENCONFIG : Querying hostname - SUCCESSFUL")
                     return resp.json()['hostname']
                 if resp.status_code != 200:
-
+                    self.log.info(
+                        f"DEVICE: {restconf_session.base_url} - RESTCONF_OPENCONFIG : Querying hostname - FAILURE")
                     # try nexus TEMPORARY
+                    self.log.info(
+                        f"DEVICE: {restconf_session.base_url} - RESTCONF_NXOS_OS_DEVICE_YANG : Querying hostname - START")
                     resp = restconf_session.get(f"{restconf_session.base_url}/Cisco-NX-OS-device:System/name")
 
                     if resp.status_code == 200:
+                        self.log.info(
+                            f"DEVICE: {restconf_session.base_url} - RESTCONF_NXOS_OS_DEVICE_YANG : Querying hostname - SUCCESSFUL")
                         return resp.json()['name']
+                    self.log.info(
+                        f"DEVICE: {restconf_session.base_url} - RESTCONF_NXOS_OS_DEVICE_YANG : Querying hostname - FAILURE")
 
             except Exception:
+
                 tb = self.get_traceback()
-                #print(tb)
+                self.log.info(
+                    f"DEVICE: {restconf_session.base_url} - RESTCONF_OPENCONFIG_AND_NXOS_OS_DEVICE_YANG_CATCHALL : Querying hostname - FAILURE - TB: {tb}")
 
         else:
-
+            self.log.info(f"DEVICE: {netmiko_session['host']} - SSH : Querying for hostname - START")
             sh_run_output = netmiko_session.send_command("sh run | inc hostname")
             #print(sh_run_output)
             hostname = re.search("hostname\s+(.*)", sh_run_output).group(1)
+            self.log.info(f"DEVICE: {netmiko_session['host']} - SSH : Querying for hostname - SUCCESSFUL")
 
             return hostname
         return None
@@ -511,21 +663,22 @@ class dyagram:
 
 def main():
 
-    #from dyagram.dyagram import dyagram
     import argparse
 
     parser = argparse.ArgumentParser()
-    #parser.add_argument('-i', dest="inventory")
-    parser.add_argument('first_arg', nargs='*')
+
+    parser.add_argument('dyagram_args', nargs='*')
 
     args = parser.parse_args()
+    if '-v' or '--verbose' in args.dyagram_args:
+        verbose = True
 
-
-    #dyagram = dyagram(args.inventory, initial=True)
-    if args.first_arg[0].lower() == "init":
+    if args.dyagram_args[0].lower() == "init":
         from dyagram.initialize import initialize
         initialize.main()
 
+
+
     if args.first_arg[0].lower() == "discover":
         print("\n\n-- DyaGram Discovering Started --\n")
         dy = dyagram()
@@ -542,7 +695,6 @@ def main():
                 try:
                     s.make_new_site(args.first_arg[2])
                 except Exception as e:
-                    #print(e)
                     print("Missing argument : <site>")
             if args.first_arg[1].lower() == "switch":
                 try:
